/**
 * @since 1.0.0
 */
import * as Either from "@fp-ts/core/Either";
import * as Option from "@fp-ts/core/Option";
import type { Predicate, Refinement } from "@fp-ts/core/Predicate";
/**
 * @since 1.0.0
 * @category symbols
 */
export declare const BrandTypeId: unique symbol;
/**
 * @since 1.0.0
 * @category symbols
 */
export type BrandTypeId = typeof BrandTypeId;
/**
 * @since 1.0.0
 * @category symbols
 */
export declare const NominalConstructorTypeId: unique symbol;
/**
 * @since 1.0.0
 * @category symbols
 */
export type NominalConstructorTypeId = typeof NominalConstructorTypeId;
/**
 * @since 1.0.0
 * @category symbols
 */
export declare const RefinedConstructorsTypeId: unique symbol;
/**
 * @since 1.0.0
 * @category symbols
 */
export type RefinedConstructorsTypeId = typeof RefinedConstructorsTypeId;
/**
 * @since 1.0.0
 * @category models
 */
export interface Brand<in out K extends string> {
    readonly [BrandTypeId]: {
        readonly [k in K]: K;
    };
}
/**
 * @since 1.0.0
 */
export declare namespace Brand {
    /**
     * @since 1.0.0
     * @category models
     */
    interface BrandErrors extends ReadonlyArray<RefinementError> {
    }
    /**
     * @since 1.0.0
     * @category models
     */
    interface RefinementError {
        readonly meta: unknown;
        readonly message: string;
    }
    /**
     * @since 1.0.0
     * @category models
     */
    type Constructor<A extends Brand<any>> = Brand.NominalConstructor<A> | Brand.RefinedConstructors<A>;
    /**
     * @since 1.0.0
     * @category models
     */
    interface NominalConstructor<in out A extends Brand<any>> {
        readonly [NominalConstructorTypeId]: NominalConstructorTypeId;
        /**
         * Constructs a branded type from a value of type `A`.
         */
        (args: Omit<Brand.Unbranded<A>, BrandTypeId>): A;
    }
    /**
     * @since 1.0.0
     * @category models
     */
    interface RefinedConstructors<in out A extends Brand<any>> {
        readonly [RefinedConstructorsTypeId]: RefinedConstructorsTypeId;
        /**
         * Constructs a branded type from a value of type `A`, throwing an error if
         * the provided `A` is not valid.
         */
        of: (args: Brand.Unbranded<A>) => A;
        /**
         * Constructs a branded type from a value of type `A`, returning `Some<A>`
         * if the provided `A` is valid, `None` otherwise.
         */
        option: (args: Brand.Unbranded<A>) => Option.Option<A>;
        /**
         * Constructs a branded type from a value of type `A`, returning `Right<A>`
         * if the provided `A` is valid, `Left<BrandError>` otherwise.
         */
        either: (args: Brand.Unbranded<A>) => Either.Either<Brand.BrandErrors, A>;
        /**
         * Attempts to refine the provided value of type `A`, returning `true` if
         * the provided `A` is valid, `false` otherwise.
         */
        refine: Refinement<Brand.Unbranded<A>, Brand.Unbranded<A> & A>;
    }
    /**
     * A utility type to extract a branded type from a `Brand.Constructor`.
     *
     * @since 1.0.0
     * @category models
     */
    type FromConstructor<A> = A extends Brand.Constructor<infer B> ? B : never;
    /**
     * A utility type to extract the value type from a brand.
     *
     * @since 1.0.0
     * @category models
     */
    type Unbranded<P> = P extends infer Q & Brands<P> ? Q : P;
    /**
     * A utility type to extract the brands from a branded type.
     *
     * @since 1.0.0
     * @category models
     */
    type Brands<P> = P extends Brand<any> ? Brand.UnionToIntersection<{
        [k in keyof P[BrandTypeId]]: k extends string ? Brand<k> : never;
    }[keyof P[BrandTypeId]]> : never;
    /**
     * @since 1.0.0
     * @category models
     */
    type EnsureCommonBase<Brands extends readonly [Brand.Constructor<any>, ...Array<Brand.Constructor<any>>]> = {
        [B in keyof Brands]: Brand.Unbranded<Brand.FromConstructor<Brands[0]>> extends Brand.Unbranded<Brand.FromConstructor<Brands[B]>> ? Brand.Unbranded<Brand.FromConstructor<Brands[B]>> extends Brand.Unbranded<Brand.FromConstructor<Brands[0]>> ? Brands[B] : Brands[B] : "ERROR: All brands should have the same base type";
    };
    /**
     * @since 1.0.0
     * @category models
     */
    type UnionToIntersection<T> = (T extends any ? (x: T) => any : never) extends (x: infer R) => any ? R : never;
}
/**
 * @since 1.0.0
 * @category constructors
 */
export declare const error: (message: string, meta?: unknown) => Brand.BrandErrors;
/**
 * @since 1.0.0
 * @category constructors
 */
export declare const errors: (...errors: Array<Brand.BrandErrors>) => Brand.BrandErrors;
/**
 * Returns `true` if the provided `Brand` is nominal, `false` otherwise.
 *
 * @since 1.0.0
 * @category refinements
 */
export declare const isNominal: <A extends Brand<any>>(u: Brand.Constructor<A>) => u is Brand.NominalConstructor<A>;
/**
 * Returns `true` if the provided `Brand` is refined, `false` otherwise.
 *
 * @since 1.0.0
 * @category refinements
 */
export declare const isRefined: <A extends Brand<any>>(u: Brand.Constructor<A>) => u is Brand.RefinedConstructors<A>;
/**
 * @since 1.0.0
 * @category constructors
 */
export declare const nominal: <A extends Brand<any>>() => Brand.NominalConstructor<A>;
/**
 * @since 1.0.0
 * @category constructors
 */
export declare const refined: <A extends Brand<any>>(refinement: Predicate<Brand.Unbranded<A>>, onFailure: (a: Brand.Unbranded<A>) => Brand.BrandErrors) => Brand.RefinedConstructors<A>;
/**
 * Composes two brands together to form a single branded type.
 *
 * @since 1.0.0
 * @category mutations
 */
export declare const all: <Brands extends readonly [Brand.Constructor<any>, ...Brand.Constructor<any>[]]>(...brands: Brand.EnsureCommonBase<Brands>) => Brand.RefinedConstructors<Brand.UnionToIntersection<{ [B in keyof Brands]: Brand.FromConstructor<Brands[B]>; }[number]> extends infer X extends Brand<any> ? X : Brand<any>>;
//# sourceMappingURL=Brand.d.ts.map