/**
 * @since 1.0.0
 */
import * as Either from "@fp-ts/core/Either";
import { identity, pipe } from "@fp-ts/core/Function";
import * as Option from "@fp-ts/core/Option";
import * as ReadonlyArray from "@fp-ts/core/ReadonlyArray";
/**
 * @since 1.0.0
 * @category symbols
 */
export const BrandTypeId = /*#__PURE__*/Symbol.for("@fp-ts/data/Brand");
/**
 * @since 1.0.0
 * @category symbols
 */
export const NominalConstructorTypeId = /*#__PURE__*/Symbol.for("@fp-ts/data/Brand/Nominal");
/**
 * @since 1.0.0
 * @category symbols
 */
export const RefinedConstructorsTypeId = /*#__PURE__*/Symbol.for("@fp-ts/data/Brand/Refined");
/**
 * @since 1.0.0
 * @category constructors
 */
export const error = (message, meta) => [{
  message,
  meta
}];
/**
 * @since 1.0.0
 * @category constructors
 */
export const errors = (...errors) => ReadonlyArray.flatten(errors);
/**
 * Returns `true` if the provided `Brand` is nominal, `false` otherwise.
 *
 * @since 1.0.0
 * @category refinements
 */
export const isNominal = u => NominalConstructorTypeId in u;
/**
 * Returns `true` if the provided `Brand` is refined, `false` otherwise.
 *
 * @since 1.0.0
 * @category refinements
 */
export const isRefined = u => RefinedConstructorsTypeId in u;
/**
 * @since 1.0.0
 * @category constructors
 */
export const nominal = () =>
// @ts-expect-error
Object.assign(args => args, {
  [NominalConstructorTypeId]: NominalConstructorTypeId
});
/**
 * @since 1.0.0
 * @category constructors
 */
export const refined = (refinement, onFailure) => {
  const either = args => refinement(args) ? Either.right(args) : Either.left(onFailure(args));
  return {
    [RefinedConstructorsTypeId]: RefinedConstructorsTypeId,
    of: args => pipe(either(args), Either.getOrThrow(identity)),
    option: args => Option.fromEither(either(args)),
    either,
    refine: args => Either.isRight(either(args))
  };
};
/**
 * Composes two brands together to form a single branded type.
 *
 * @since 1.0.0
 * @category mutations
 */
export const all = (...brands) => {
  const either = args => {
    let result = Either.right(args);
    for (const brand of brands) {
      if (isRefined(brand)) {
        const nextResult = brand.either(args);
        if (Either.isLeft(result) && Either.isLeft(nextResult)) {
          result = Either.left([...result.left, ...nextResult.left]);
        } else {
          result = Either.isLeft(result) ? result : nextResult;
        }
      }
    }
    return result;
  };
  return {
    [RefinedConstructorsTypeId]: RefinedConstructorsTypeId,
    of: args => pipe(either(args), Either.getOrThrow(identity)),
    option: args => Option.fromEither(either(args)),
    either,
    refine: args => Either.isRight(either(args))
  };
};
//# sourceMappingURL=Brand.mjs.map